@@ -79,6 +79,9 @@
 #define DEFAULT_CARD_NUM 0
 
 #define VERSION_STRING "malsync version 2.1.1"
+#define AG_EXPANSION_VERSION_NONE (-1)
+#define AVANTGO_VERSION_DATABASE_NAME   "AvGoVersion"
+#define AVANTGO_AGVERSION_DATABASE_NAME "AGVersion"
 
 static char *device = "/dev/pilot";
 static char *httpProxy      = NULL;
@@ -137,6 +140,8 @@ static int threeone = 0;
 static int lowres = 0;
 
 /* ----------------------------------------------------------------------------*/
+static void fill_in_versioninfo(int sd, AGExpansionVersion *vi);
+static void ClientVersionReadData(ClientVersion *cversion, uint32 len, uint8 *buf);
 static long openUserConfigDatabase(int *threeone);
 static int32 readDeviceUserConfig32(int userConfigDBHandle, AGUserConfig **deviceUserConfig);
 static int32 readDeviceUserConfig31(int userConfigDBHandle, AGUserConfig **deviceUserConfig);
@@ -252,8 +257,13 @@ void syncInfoFree(PalmSyncInfo * pInfo)
         if (NULL != pInfo->userConfig)
             AGUserConfigFree(pInfo->userConfig);
 
+#ifdef PILOT_LINK_0_12
+        if (NULL != pInfo->pi_buf)
+            pi_buffer_free(pInfo->pi_buf);
+#else
         if (NULL != pInfo->pilot_buffer)
             free(pInfo->pilot_buffer);
+#endif
 
         if (NULL != pInfo->commandProcessor)
             AGCommandProcessorFree(pInfo->commandProcessor);
@@ -275,10 +285,16 @@ PalmSyncInfo * syncInfoNew()
 
         bzero(pInfo, sizeof(PalmSyncInfo));
 
+#ifdef PILOT_LINK_0_12
+        pInfo->pi_buf = pi_buffer_new(pbs);
+        if (NULL == pInfo->pi_buf)
+            goto fail;
+#else
         pInfo->pilot_buffer_size    = pbs;
         pInfo->pilot_buffer         = (uint8 *)malloc(pbs);
         if (NULL == pInfo->pilot_buffer)
             goto fail;
+#endif
 
         /* Allocate the platform calls record. */
         pInfo->platform = (AGPlatformCalls *)malloc(sizeof(AGPlatformCalls));
@@ -324,8 +340,12 @@ int16 readInt16(uint8 * p)
 /*---------------------------------------------------------------------------*/
 static void
 readAndUseDeviceInfoDatabase(AGDeviceInfo * devInfo,
+#ifdef PILOT_LINK_0_12
+                             pi_buffer_t *pi_buf)
+#else
                              uint8 *dev_db_info_buffer,
                              uint32 dev_db_info_buffer_size)
+#endif
 {
     int database_id = 0;
     long result;
@@ -357,7 +377,7 @@ readAndUseDeviceInfoDatabase(AGDeviceInfo * devInfo,
 
 #ifdef PILOT_LINK_0_12
         rc = dlp_ReadRecordByIndex(sd, database_id, 0, 
-                                   (void *)dev_db_info_buffer,
+                                   pi_buf,
                                    &id, &attr, &cat);
 #else
         rc = dlp_ReadRecordByIndex(sd, database_id, 0,
@@ -368,7 +388,11 @@ readAndUseDeviceInfoDatabase(AGDeviceInfo * devInfo,
 
 
         if (rc >= 0) {
+#ifdef PILOT_LINK_0_12
+            uint8 *p = pi_buf->data;
+#else
             uint8 *p = dev_db_info_buffer;
+#endif
             int16 dev_db_info_version = readInt16(p);
             p+=sizeof(int16);
             devInfo->colorDepth = readInt32(p);
@@ -457,8 +481,12 @@ static int readDeviceInfo(PalmSyncInfo * pInfo)
     }
 
     readAndUseDeviceInfoDatabase(devInfo,
+#ifdef PILOT_LINK_0_12
+                                 pInfo->pi_buf);
+#else
                                  pInfo->pilot_buffer,
                                  pInfo->pilot_buffer_size);
+#endif
 
     /* Override the color depth if the user wants low res images. */
     if (lowres) {
@@ -634,15 +662,20 @@ static int32
 readDeviceUserConfig32(int userConfigDBHandle, AGUserConfig **deviceUserConfig)
 {
     recordid_t id;
+#ifdef PILOT_LINK_0_12
+    pi_buffer_t *pi_buf;
+#else
     int bufferSize = BUFFERSIZE;
+    uint8 buffer[BUFFERSIZE];
+#endif
     int attr = 0;
     int cat  = 0;
     int rc;
-    uint8 buffer[BUFFERSIZE];
     AGBufferReader * r = NULL;
 
 #ifdef PILOT_LINK_0_12
-    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
+    pi_buf = pi_buffer_new(BUFFERSIZE);
+    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, pi_buf,
                                &id, &attr, &cat);
 #else
     rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
@@ -652,18 +685,32 @@ readDeviceUserConfig32(int userConfigDBHandle, AGUserConfig **deviceUserConfig)
     if (rc < 0) {
 #ifdef _DEBUG
         printf("readDeviceUserConfig: dlp_ReadRecordByIndex , err = %d\n", rc);
+#endif
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
 #endif
         return 0;
     }
 
+#ifdef PILOT_LINK_0_12
+    r = AGBufferReaderNew(pi_buf->data);
+#else
     r = AGBufferReaderNew(buffer);
+#endif
     if (r) {
         *deviceUserConfig = AGUserConfigNew();
         AGUserConfigReadData(*deviceUserConfig, (AGReader*)r);
         AGBufferReaderFree(r);
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
+#endif
         return id;
-    } else
+    } else {
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
+#endif
         return 0;
+    }
 }
 #define BUFFERSIZE 0xFFFF
 /*---------------------------------------------------------------------------*/
@@ -671,15 +718,20 @@ static int32 readDeviceUserConfig31(int userConfigDBHandle,
                                     AGUserConfig **deviceUserConfig)
 {
     recordid_t id;
+#ifdef PILOT_LINK_0_12
+    pi_buffer_t *pi_buf;
+#else
     int bufferSize = BUFFERSIZE;
+    uint8 buffer[BUFFERSIZE];
+#endif
     int attr = 0;
     int cat  = 0;
     int rc;
-    uint8 buffer[BUFFERSIZE];
     AGBufferReader * r = NULL;
 
 #ifdef PILOT_LINK_0_12
-    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
+    pi_buf = pi_buffer_new(BUFFERSIZE);
+    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, pi_buf,
                                &id, &attr, &cat);
 #else
     rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
@@ -690,18 +742,32 @@ static int32 readDeviceUserConfig31(int userConfigDBHandle,
 #ifdef _DEBUG
         printf("readDeviceUserConfig: dlp_ReadRecordByIndex , err = %d\n",
                rc);
+#endif
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
 #endif
         return 0;
     }
 
+#ifdef PILOT_LINK_0_12
+    r = AGBufferReaderNew(pi_buf->data);
+#else
     r = AGBufferReaderNew(buffer);
+#endif
     if (r) {
         *deviceUserConfig = AGUserConfigNew();
         MAL31ReadUserData(*deviceUserConfig, (AGReader*)r);
         AGBufferReaderFree(r);
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
+#endif
         return id;
-    } else
+    } else {
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
+#endif
         return 0;
+    }
 }
 /*---------------------------------------------------------------------------*/
 static int32 readDeviceUserConfig(int userConfigDBHandle,
@@ -723,14 +789,21 @@ static void writeDeviceUserConfig(int userConfigDBHandle,
 {
 
     recordid_t id;
+#ifndef PILOT_LINK_0_12
     int bufferSize = BUFFERSIZE;
+    uint8 buffer[BUFFERSIZE];
+#endif
     int attr = 0;
     int cat  = 0;
-    uint8 buffer[BUFFERSIZE];
     AGBufferWriter * w = NULL;
     w = AGBufferWriterNew(0);
     if (w) {
         long result;
+#ifdef PILOT_LINK_0_12
+        pi_buffer_t *pi_buf;
+
+        pi_buf = pi_buffer_new(BUFFERSIZE);
+#endif
 
         if (threeone) {
             MAL31WriteUserData(deviceUserConfig, (AGWriter*)w);
@@ -739,7 +812,7 @@ static void writeDeviceUserConfig(int userConfigDBHandle,
         }
 
 #ifdef PILOT_LINK_0_12
-        result = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
+        result = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, pi_buf,
                                        &id, &attr, &cat);
 #else
         result = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
@@ -755,6 +828,9 @@ static void writeDeviceUserConfig(int userConfigDBHandle,
                                   &id);
         AGBufferWriterFree(w);
 
+#ifdef PILOT_LINK_0_12
+        pi_buffer_free(pi_buf);
+#endif
     }
 }
 
@@ -1022,16 +1098,18 @@ static int32 getRecordBase(PalmSyncInfo * pInfo, AGBool modonly,
     int32 result;
     int att = 0;
     int cat = 0;
+#ifndef PILOT_LINK_0_12
     int size = pInfo->pilot_buffer_size;
+#endif
     int idx   = pInfo->pilot_RecIndex++;
 
     result = (modonly) ?
 #ifdef PILOT_LINK_0_12
-        dlp_ReadNextModifiedRec (sd, pInfo->pilot_rHandle, pInfo->pilot_buffer,
+        dlp_ReadNextModifiedRec (sd, pInfo->pilot_rHandle, pInfo->pi_buf,
                                  &pInfo->id, &idx, &att, &cat)
         :
         dlp_ReadRecordByIndex(sd, pInfo->pilot_rHandle, idx,
-                              (void *)pInfo->pilot_buffer,
+                              (void *)pInfo->pi_buf,
                               &pInfo->id, &att, &cat);
 #else
         dlp_ReadNextModifiedRec (sd, pInfo->pilot_rHandle, pInfo->pilot_buffer,
@@ -1056,7 +1134,11 @@ static int32 getRecordBase(PalmSyncInfo * pInfo, AGBool modonly,
     }
     pInfo->record = AGRecordInit(pInfo->record, pInfo->id,
                                  AGPalmPilotAttribsToMALMod((uint8)att),
+#ifdef PILOT_LINK_0_12
+                                 pInfo->pi_buf->used, pInfo->pi_buf->data, 0, NULL);
+#else
                                  size, pInfo->pilot_buffer, 0, NULL);
+#endif
 
     *record = pInfo->record;
     return AGCLIENT_CONTINUE;
@@ -1207,6 +1289,16 @@ static AGBool doClientProcessorLoop(PalmSyncInfo * pInfo, AGNetCtx *ctx)
                 AGDeviceInfoFree(pInfo->deviceInfo);
                 return FALSE;
             }
+
+            pInfo->clientProcessor->version_info =
+                malloc(sizeof(AGExpansionVersion));
+            if (NULL == pInfo->clientProcessor->version_info) {
+                jp_logf(JP_LOG_FATAL,"%s\n", "Error allocating memory for version info");
+                return -1;
+            }
+
+            fill_in_versioninfo(sd, pInfo->clientProcessor->version_info);
+
             AGClientProcessorSetBufferServerCommands(pInfo->clientProcessor,
                                                      FALSE);
 
@@ -1326,3 +1418,88 @@ int malsync (int sd_hack, PalmSyncInfo *pInfo) {
 
     return 0;
 }
+
+static void ClientVersionReadData(ClientVersion *cversion, uint32 len, uint8 *buf) {
+    int32 version;
+    AGBufferReader r;
+
+    bzero(cversion, sizeof(ClientVersion));
+
+    AGBufferReaderInit(&r, buf);
+    AGReadInt32((AGReader *)&r); /* old data compatibility field */
+    if (len > sizeof(int32)) {
+        version = AGReadInt32((AGReader *)&r);
+        cversion->dataCompatibility = AGReadInt32((AGReader *)&r);
+        cversion->major = AGReadInt32((AGReader *)&r);
+        cversion->minor = AGReadInt32((AGReader *)&r);
+        cversion->buildno = AGReadInt32((AGReader *)&r);
+    } else {
+        cversion->dataCompatibility = 1;
+    }
+    AGBufferReaderFinalize(&r);
+}
+
+/*----------------------------------------------------------------------------*/
+static void
+fill_in_versioninfo(int sd, AGExpansionVersion *vi)
+{
+    int database_id = 0;
+    long result = -1;
+
+    /* Assume failure of device portion and write reasonable values. */
+    vi->device_major = AG_EXPANSION_VERSION_NONE;
+    vi->device_minor = AG_EXPANSION_VERSION_NONE;
+    vi->device_build = AG_EXPANSION_VERSION_NONE;
+    vi->conn_agent_major = 1;
+    vi->conn_agent_minor = 0;
+    vi->conn_agent_build = 0;
+
+    result = dlp_OpenDB(sd,
+            DEFAULT_CARD_NUM,
+            dlpOpenRead,
+            AVANTGO_AGVERSION_DATABASE_NAME,
+            &database_id);
+
+    /* 4.x and earlier client has AvGoVersion db, try that */
+    if (result < 0)
+        result = dlp_OpenDB(sd,
+                DEFAULT_CARD_NUM,
+                dlpOpenRead,
+                AVANTGO_VERSION_DATABASE_NAME,
+                &database_id);
+
+    if (result >= 0) {
+        long syncReadResult;
+        int attr = 0;
+        int cat  = 0;
+        recordid_t id;
+        int rc;
+        pi_buffer_t *version_buffer = pi_buffer_new(0x1000);
+
+        rc = dlp_ReadRecordByIndex(sd, database_id,
+                0,
+                version_buffer,
+                &id,
+                &attr,
+                &cat);
+
+        if (rc >= 0) {
+            ClientVersion cv;
+            ClientVersionReadData(&cv,
+                    version_buffer->used,
+                    version_buffer->data);
+            vi->device_major = cv.major;
+            vi->device_minor = cv.minor;
+            vi->device_build = cv.buildno;
+
+            vi->conn_agent_major = cv.major;
+            vi->conn_agent_minor = cv.minor;
+            vi->conn_agent_build = cv.buildno;
+        }
+
+        pi_buffer_free(version_buffer);
+
+        dlp_CloseDB(sd, database_id);
+
+    }
+}
